/*
 * Copyright (c) 2014, 2015, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.r.runtime.conn;

import static com.oracle.truffle.r.runtime.conn.ConnectionSupport.*;

import java.io.*;
import java.nio.*;
import java.util.zip.*;

import com.oracle.truffle.api.CompilerDirectives.*;
import com.oracle.truffle.r.runtime.*;
import com.oracle.truffle.r.runtime.data.model.*;

public class FileConnections {
    /**
     * Base class for all modes of file connections.
     *
     */
    public static class FileRConnection extends BasePathRConnection {

        public FileRConnection(String path, String modeString) throws IOException {
            super(checkTemp(path), ConnectionClass.File, modeString);
            openNonLazyConnection();
        }

        private static String checkTemp(String path) {
            if (path.length() == 0) {
                return TempPathName.createNonExistingFilePath("Rf", TempPathName.tempDirPath(), "");
            } else {
                return path;
            }
        }

        @Override
        protected void createDelegateConnection() throws IOException {
            DelegateRConnection delegate = null;
            switch (getOpenMode().abstractOpenMode) {
                case Read:
                    delegate = new FileReadTextRConnection(this);
                    break;
                case Write:
                    delegate = new FileWriteTextRConnection(this, false);
                    break;
                case Append:
                    delegate = new FileWriteTextRConnection(this, true);
                    break;
                case ReadBinary:
                    delegate = new FileReadBinaryRConnection(this);
                    break;
                case WriteBinary:
                    delegate = new FileWriteBinaryConnection(this, false);
                    break;
                case ReadWriteTrunc:
                case ReadWriteTruncBinary:
                    delegate = new FileReadWriteConnection(this);
                    break;
                default:
                    throw RError.nyi(RError.NO_NODE, "open mode: " + getOpenMode());
            }
            setDelegate(delegate);
        }
    }

    static class FileReadTextRConnection extends DelegateReadRConnection implements ReadWriteHelper {
        private InputStream inputStream;

        FileReadTextRConnection(BasePathRConnection base) throws IOException {
            super(base);
            inputStream = new BufferedInputStream(new FileInputStream(base.path));
            // can be compressed - check for it
            inputStream.mark(2);
            int byte1 = inputStream.read();
            if (byte1 == -1) {
                inputStream.reset();
            } else {
                int byte2 = inputStream.read();
                if (byte2 == -1) {
                    inputStream.reset();
                } else {
                    if (byte1 == (GZIPInputStream.GZIP_MAGIC & 0x000000FF) && byte2 == (GZIPInputStream.GZIP_MAGIC >> 8)) {
                        inputStream.close();
                        inputStream = new GZIPInputStream(new FileInputStream(base.path), GZIPConnections.GZIP_BUFFER_SIZE);
                    } else {
                        inputStream.reset();
                    }
                }
            }
        }

        @Override
        public int readBin(ByteBuffer buffer) throws IOException {
            throw RError.error(RError.NO_NODE, RError.Message.ONLY_READ_BINARY_CONNECTION);
        }

        @Override
        public byte[] readBinChars() throws IOException {
            throw RError.error(RError.NO_NODE, RError.Message.ONLY_READ_BINARY_CONNECTION);
        }

        @TruffleBoundary
        @Override
        public String[] readLinesInternal(int n) throws IOException {
            return readLinesHelper(inputStream, n);
        }

        @Override
        public String readChar(int nchars, boolean useBytes) throws IOException {
            return readCharHelper(nchars, inputStream, useBytes);
        }

        @Override
        public InputStream getInputStream() throws IOException {
            return inputStream;
        }

        @Override
        public void closeAndDestroy() throws IOException {
            base.closed = true;
            close();
        }

        @Override
        public void close() throws IOException {
            inputStream.close();
        }

    }

    private static class FileWriteTextRConnection extends DelegateWriteRConnection implements ReadWriteHelper {
        private BufferedOutputStream outputStream;

        FileWriteTextRConnection(FileRConnection base, boolean append) throws IOException {
            super(base);
            outputStream = new BufferedOutputStream(new FileOutputStream(base.path, append));
        }

        @Override
        public void writeChar(String s, int pad, String eos, boolean useBytes) throws IOException {
            writeCharHelper(outputStream, s, pad, eos);
        }

        @Override
        public void writeBin(ByteBuffer buffer) throws IOException {
            throw RError.error(RError.NO_NODE, RError.Message.ONLY_WRITE_BINARY_CONNECTION);
        }

        @Override
        public void writeLines(RAbstractStringVector lines, String sep, boolean useBytes) throws IOException {
            writeLinesHelper(outputStream, lines, sep);
            flush();
        }

        @Override
        public void writeString(String s, boolean nl) throws IOException {
            writeStringHelper(outputStream, s, nl);
        }

        @Override
        public OutputStream getOutputStream() throws IOException {
            return outputStream;
        }

        @Override
        public void closeAndDestroy() throws IOException {
            base.closed = true;
            close();
        }

        @Override
        public void close() throws IOException {
            outputStream.close();
        }

        @Override
        public void flush() throws IOException {
            outputStream.flush();
        }
    }

    private static class FileReadBinaryRConnection extends DelegateReadRConnection implements ReadWriteHelper {
        private FileInputStream inputStream;

        FileReadBinaryRConnection(FileRConnection base) throws IOException {
            super(base);
            inputStream = new FileInputStream(base.path);
        }

        @Override
        public String readChar(int nchars, boolean useBytes) throws IOException {
            return readCharHelper(nchars, inputStream, useBytes);
        }

        @Override
        public int readBin(ByteBuffer buffer) throws IOException {
            return inputStream.getChannel().read(buffer);
        }

        @Override
        public byte[] readBinChars() throws IOException {
            return readBinCharsHelper(inputStream);
        }

        @TruffleBoundary
        @Override
        public String[] readLinesInternal(int n) throws IOException {
            return readLinesHelper(inputStream, n);
        }

        @Override
        public InputStream getInputStream() throws IOException {
            return inputStream;
        }

        @Override
        public void closeAndDestroy() throws IOException {
            base.closed = true;
            close();
        }

        @Override
        public void close() throws IOException {
            inputStream.close();
        }

    }

    private static class FileWriteBinaryConnection extends DelegateWriteRConnection implements ReadWriteHelper {
        private FileOutputStream outputStream;

        FileWriteBinaryConnection(FileRConnection base, boolean append) throws IOException {
            super(base);
            outputStream = new FileOutputStream(base.path, append);
        }

        @Override
        public void writeChar(String s, int pad, String eos, boolean useBytes) throws IOException {
            writeCharHelper(outputStream, s, pad, eos);
        }

        @Override
        public void writeBin(ByteBuffer buffer) throws IOException {
            outputStream.getChannel().write(buffer);
        }

        @Override
        public OutputStream getOutputStream() throws IOException {
            return outputStream;
        }

        @Override
        public void closeAndDestroy() throws IOException {
            base.closed = true;
            close();
        }

        @Override
        public void close() throws IOException {
            flush();
            outputStream.close();
        }

        @Override
        public void writeLines(RAbstractStringVector lines, String sep, boolean useBytes) throws IOException {
            for (int i = 0; i < lines.getLength(); i++) {
                String line = lines.getDataAt(i);
                outputStream.write(line.getBytes());
                outputStream.write(sep.getBytes());
            }
        }

        @Override
        public void writeString(String s, boolean nl) throws IOException {
            writeStringHelper(outputStream, s, nl);
        }

        @Override
        public void flush() throws IOException {
            outputStream.flush();
        }

    }

    private static class FileReadWriteConnection extends DelegateReadWriteRConnection implements ReadWriteHelper {
        /*
         * This does not fit in well as it does not support streams so most of the helpers are
         * useless. Life would be a little better if we converted everything to channels, as it does
         * support those. This is a minimal implementation to support one specific use in package
         * installation (write only).
         * 
         * N.B. R mandates separate "position" offsets for reading and writing (pain). This code is
         * pessimistic and assumes interleaved reads and writes, so does a lot of probably redundant
         * seeking. It could be optimized.
         */
        private final RandomAccessFile raf;
        private long readOffset;
        private long writeOffset;
        private SeekRWMode lastMode = SeekRWMode.READ;

        FileReadWriteConnection(FileRConnection base) throws IOException {
            super(base);
            OpenMode openMode = base.getOpenMode();
            String rafMode = null;
            switch (openMode.abstractOpenMode) {
                case ReadWriteTrunc:
                case ReadWriteTruncBinary:
                    rafMode = "rw";
                    break;
                default:
                    throw RInternalError.shouldNotReachHere();
            }
            raf = new RandomAccessFile(base.path, rafMode);
        }

        @Override
        public int getc() throws IOException {
            raf.seek(readOffset);
            int value = raf.read();
            readOffset++;
            return value;
        }

        @Override
        public boolean isSeekable() {
            return true;
        }

        @Override
        public long seek(long offset, SeekMode seekMode, SeekRWMode seekRWMode) throws IOException {
            long result = raf.getFilePointer();
            if (seekMode != SeekMode.START) {
                throw RError.nyi(null, "seek mode");
            }
            switch (seekRWMode) {
                case LAST:
                    if (lastMode == SeekRWMode.READ) {
                        readOffset = offset;
                    } else {
                        writeOffset = offset;
                    }
                    break;
                case READ:
                    readOffset = offset;
                    break;
                case WRITE:
                    writeOffset = offset;
                    break;
            }
            return result;
        }

        @Override
        public String[] readLinesInternal(int n) throws IOException {
            throw RInternalError.unimplemented();
        }

        @Override
        public InputStream getInputStream() throws IOException {
            throw RInternalError.unimplemented();
        }

        @Override
        public OutputStream getOutputStream() throws IOException {
            throw RInternalError.unimplemented();
        }

        @Override
        public void closeAndDestroy() throws IOException {
            base.closed = true;
            close();
        }

        @Override
        public void close() throws IOException {
            raf.close();
        }

        @Override
        public void writeLines(RAbstractStringVector lines, String sep, boolean useBytes) throws IOException {
            // TODO encodings
            raf.seek(writeOffset);
            byte[] sepData = sep.getBytes();
            for (int i = 0; i < lines.getLength(); i++) {
                byte[] data = lines.getDataAt(i).getBytes();
                raf.write(data);
                raf.write(sepData);
            }
            writeOffset = raf.getFilePointer();
            lastMode = SeekRWMode.WRITE;
        }

        @Override
        public void flush() throws IOException {
        }

        @Override
        public void writeString(String s, boolean nl) throws IOException {
            throw RInternalError.unimplemented();
        }

        @Override
        public void writeChar(String s, int pad, String eos, boolean useBytes) throws IOException {
            throw RInternalError.unimplemented();
        }

        @Override
        public String readChar(int nchars, boolean useBytes) throws IOException {
            throw RInternalError.unimplemented();
        }

        @Override
        public void writeBin(ByteBuffer buffer) throws IOException {
            throw RInternalError.unimplemented();
        }

        @Override
        public int readBin(ByteBuffer buffer) throws IOException {
            throw RInternalError.unimplemented();
        }

        @Override
        public byte[] readBinChars() throws IOException {
            throw RInternalError.unimplemented();
        }
    }

}
